import requests


class Stream(object):
    def __init__(self, host, url, listener, stream, token, protocol='https'):
        self.protocol = protocol
        self.host = host
        self.url = url
        self.stream = stream
        self.token = token
        self.listener = listener

        self.streaming = False

    def go(self):
        url = '%s://%s%s' % (self.protocol, self.host, self.url)

        while True:
            response = requests.get(
                url,
                {'stream': self.stream},
                stream=True,
                headers={
                    'Authorization': 'Token %s' % self.token
                }
            )

            if not response.ok:
                print 'Something went wrong!'
                print 'Status:', response.status_code
                print response.content
                return

            self.streaming = True
            self._consume(response)

            print 'Stream dropped: reconnecting'

    def _consume(self, resp):
        buf = resp.raw

        while not resp.raw.closed:
            line = buf.readline().strip()

            if not line:
                self.listener.on_heartbeat()
            elif line.isdigit():
                length = int(line)
                header = buf.read(length)
                length = int(buf.readline().strip())
                body = buf.read(length - 2)
                self.listener.on_message(header, body)

        print 'Connection closed.'
        print resp.raw.read()



"""API object handling communication with asmaster, for use in manager mode."""

from __future__ import absolute_import

import requests

from . import utils
from .conf import settings


class AsmasterApi(object):
    """Primary object that pushes requests into a distinct stream thread."""

    def __init__(self, timeout):
        self.timeout = timeout
        self.host = settings.get('hosts', 'asmaster_host')
        self.token = settings.get('auth', 'token')

        self.list_services_endpoint = self._build_endpoint('list_services')
        self.list_subscriptions_endpoint = self._build_endpoint('list_subscriptions')
        self.subscribe_account_endpoint = self._build_endpoint('subscribe_account')
        self.perform_2fa_challenge_endpoint = self._build_endpoint('perform_2fa_challenge')
        self.submit_2fa_challenge_endpoint = self._build_endpoint('submit_2fa_challenge')
        self.list_devices_endpoint = self._build_endpoint('list_devices')
        self.subscribe_device_endpoint = self._build_endpoint('subscribe_device')
        self.resubscribe_account_endpoint = self._build_endpoint('resubscribe_account')
        self.unsubscribe_device_endpoint = self._build_endpoint('unsubscribe_device')
        self.unsubscribe_account_endpoint = self._build_endpoint('unsubscribe_account')
        self.reset_subscription_since_endpoint = self._build_endpoint('reset_subscription_since')

        self.services = {}

    def _build_endpoint(self, endpoint_name):
        """Generate an enpoint url from a setting name.

        Args:
            endpoint_name(str): setting name for the enpoint to build

        Returns:
            (str) url enpoint
        """
        endpoint_relative = settings.get('asmaster_endpoints', endpoint_name)
        return '%s%s' % (self.host, endpoint_relative)

    @property
    def token_header(self):
        return {
            'Authorization': 'Token %s' % self.token,
        }

    @staticmethod
    def _parse_endpoint(endpoint):
        """Expect endpoint to be dictionary containing `protocol`, `host` and `uri` keys."""
        return "{protocol}://{host}{uri}".format(**endpoint)

    def _set_endpoints(self, info):
        self.stream_endpoints = info['stream_endpoints']

    def _set_allowed_services_and_actions(self, services):
        """Expect services to be a list of service dictionaries, each with `name` and `actions` keys."""
        for service in services:
            self.services[service['name']] = {}

            for action in service['actions']:
                name = action.pop('name')
                self.services[service['name']][name] = action

    def setup(self):
        info = self.list_services()
        self._set_endpoints(info)
        self.retrieval_protocol = None
        self._set_allowed_services_and_actions(info['services'])

    def allowed_services(self):
        return self.services.keys()

    def allowed_actions(self, service_name):
        return self.services[service_name].keys()

    def list_services(self):
        """Asks for a list of all services."""
        return self._perform_post_request(self.list_services_endpoint, {}, self.token_header)

    def list_subscriptions(self, service):
        """Asks for a list of all subscribed accounts and devices, along with their statuses."""
        data = {
            'service': service,
        }
        return self._perform_post_request(self.list_subscriptions_endpoint, data, self.token_header)

    def subscribe_account(self, username, password, service):
        """Subscribe an account for a service.
        """
        data = {
            'service': service,
            'username': username,
            'password': password,
        }

        return self._perform_post_request(self.subscribe_account_endpoint, data, self.token_header)

    def perform_2fa_challenge(self, account_id, device_id):
        data = {
            'account_id': account_id,
            'device_id': device_id,
        }
        return self._perform_post_request(self.perform_2fa_challenge_endpoint, data, self.token_header)

    def submit_2fa_challenge(self, account_id, code):
        data = {
            'account_id': account_id,
            'code': code,
        }
        return self._perform_post_request(self.submit_2fa_challenge_endpoint, data, self.token_header)

    def resubscribe_account(self, account_id, password):
        data = {
            'account_id': account_id,
            'password': password,
        }

        return self._perform_post_request(self.resubscribe_account_endpoint, data, self.token_header)

    def unsubscribe_account(self, account_id):
        data = {
            'account_id': account_id,
        }

        return self._perform_post_request(self.unsubscribe_account_endpoint, data, self.token_header)

    def list_devices(self, account_id):
        data = {
            'account_id': account_id,
        }

        return self._perform_post_request(self.list_devices_endpoint, data, self.token_header)

    def subscribe_device(self, account_id, device_id):
        data = {
            'account_id': account_id,
            'device_id': device_id,
        }

        return self._perform_post_request(self.subscribe_device_endpoint, data, self.token_header)

    def unsubscribe_device(self, account_id, device_id):
        data = {
            'account_id': account_id,
            'device_id': device_id,
        }

        return self._perform_post_request(self.unsubscribe_device_endpoint, data, self.token_header)

    def reset_subscription_since(self, account_id, datetime_str):
        """Handler for `--reset-subscription-since` command.

        Args:
            account_id(int): id of the account to reset.
            datetime_str(str): string representing the datetime used in the
                next poll to retrieve data since.

        Returns:
            (str) json encoded response.

        NOTES:
            We don't care about validation here, we demand the responsibility to
            the backend.
        """
        data = {
            'account_id': account_id,
            'datetime': datetime_str,
        }
        return self._perform_post_request(self.reset_subscription_since_endpoint, data, self.token_header)

    @staticmethod
    def _parse_response(response, post_request=False):
        """Treat the response from ASApi.

        The json is dumped before checking the status as even if the response is
        not properly formed we are in trouble.
        """
        try:
            data = response.json()
        except:
            msg = 'Unhandled HTTP %s response, shown truncated below:\n%s...' % (
                response.status_code, response.text[:50]
            )
            raise ValueError(msg)

        if not response.ok:
            utils.error_message_and_exit(None, data)

        if post_request and not data['success']:
            raise Exception('Asmaster Api Error: [%s]' % data['error'])

        return data

    def _perform_get_request(self, url, headers=None):
        response = requests.get(
            url,
            headers=headers,
            timeout=self.timeout,
        )
        return self._parse_response(response)

    def _perform_post_request(self, url, data, headers=None):
        response = requests.post(
            url,
            data=data,
            headers=headers,
            timeout=self.timeout,
        )
        return self._parse_response(response, post_request=True)



"""Base module for `ricloud`."""
from __future__ import absolute_import, unicode_literals

import threading
from cached_property import cached_property

from .conf import settings
from .api import Api
from .stream import Stream
from .object_store import ObjectStore
from .listener import Listener
from .handlers import RiCloudHandler


class RiCloud(object):
    """Primary class for dealing with the ricloud API/"""

    def __init__(self, api=None, listener=None):
        self.api = api or Api()
        self.api.setup()
        self.listener = listener or Listener({'__ALL__': RiCloudHandler(api=self.api)})
        self.services = self.api.allowed_services()
        if self.api.retrieval_protocol == "asstore":
            self.object_store_thread.start()
        else:
            self.stream_thread.start()

    @property
    def stream_thread_is_daemon(self):
        return True

    @cached_property
    def stream_thread(self):
        stream = settings.get('stream', 'stream_endpoint')
        token = settings.get('auth', 'token')

        def _start():
            s = Stream(
                host=self.api.stream_endpoints[0]['host'],
                url=self.api.stream_endpoints[0]['uri'],
                listener=self.listener,
                stream=stream,
                token=token,
                protocol=self.api.stream_endpoints[0]['protocol']
            )
            s.go()

        stream_thread = threading.Thread(target=_start)
        stream_thread.daemon = self.stream_thread_is_daemon
        return stream_thread

    @cached_property
    def object_store_thread(self):

        def _start():
            s = ObjectStore(
                api=self.api
            )
            s.go()
        object_store_thread = threading.Thread(target=_start)
        object_store_thread.daemon = True
        return object_store_thread



from .base import BaseClient, sync


class iCloudClient(BaseClient):
    service = 'iCloud'

    @property
    def available_data(self):
        if 'permissions' in self.ricloud.api.services['iCloud']['Fetch Data']:
            return self.ricloud.api.services['iCloud']['Fetch Data']['permissions']['data']
        else:
            return None

    @sync
    def login(self, account, password):
        return self.ricloud.api.perform_task(
            service=self.service,
            task_name='log-in',
            account=account,
            payload={
                'password': password,
            },
        )

    @sync
    def start_2fa_auth(self, account, challenge=None):
        payload = {}

        if challenge is not None:
            payload['challenge'] = challenge

        return self.ricloud.api.perform_task(
            service=self.service,
            task_name='perform-2fa-challenge',
            account=account,
            payload=payload
        )

    @sync
    def finish_2fa_auth(self, account, code):
        return self.ricloud.api.perform_task(
            service=self.service,
            task_name='submit-2fa-challenge',
            account=account,
            payload={
                'code': code,
            },
        )

    def devices(self, account):
        return self.ricloud.api.perform_task(
            service=self.service,
            task_name='list-devices',
            account=account,
            payload={},
            callback=None)

    def data(self, account, device, data, callback=None):
        return self.ricloud.api.perform_task(
            service=self.service,
            task_name='fetch-data',
            account=account,
            payload={
                'data': data,
                'device': device
            },
            callback=callback
        )

    def download_file(self, account, device, file, callback=None):
        return self.ricloud.api.perform_task(
            service=self.service,
            task_name='download-file',
            account=account,
            payload={
                'file': file,
                'device': device,
            },
            callback=callback
        )



import os
import re
import json
import shutil
import hashlib
import logging
import MySQLdb
from io import BytesIO
from cached_property import cached_property

from .api import Api
from conf import (
    OUTPUT_DIR, LISTENER_DB_HOST, LISTENER_DB_PORT, LISTENER_DB_NAME,
    LISTENER_DB_USER, LISTENER_DB_PASSWORD
)
from .ricloud import RiCloud
from .listener import Listener
from .handlers import RiCloudHandler, StreamError
from .asmaster_api import AsmasterApi


class AsmasterListener(RiCloud):
    def __init__(self, timeout):
        self.timeout = timeout
        super(AsmasterListener, self).__init__(
            listener=Listener({'__ALL__': DatabaseWrtingHandler()}),
        )

    @property
    def stream_thread_is_daemon(self):
        return False


class DatabaseWrtingHandler(RiCloudHandler):
    def __init__(self, api=None, callback=None, db=LISTENER_DB_NAME, file_location=OUTPUT_DIR):
        self.conn = MySQLdb.connect(
            host=LISTENER_DB_HOST,
            port=int(LISTENER_DB_PORT),
            user=LISTENER_DB_USER,
            passwd=LISTENER_DB_PASSWORD,
            db=db,
        )
        self.file_location = file_location
        super(DatabaseWrtingHandler, self).__init__(api, callback)

    def on_complete_message(self, header, stream):
        if header['type'] == 'system':
            body = stream.read()
            json_body = json.loads(body)
            try:
                code = json_body['code'][:200]
            except KeyError:
                code = None

            cursor = self.conn.cursor()

            cursor.execute("""
                INSERT INTO system (`received`, `headers`, `body`, `message`, `code`)
                VALUES (NOW(), %s, %s, %s, %s)""", (
                    json.dumps(header),
                    body,
                    json_body['message'][:200],
                    code,
                )
            )
            self.conn.commit()

            if "error" in body:
                raise StreamError(body)

        elif header['type'] in {'fetch-data', 'message'}:
            if header['type'] == 'fetch-data':
                table = 'feed'
            else:
                table = header['type']

            cursor = self.conn.cursor()
            cursor.execute("""
                INSERT INTO {} (`service`, `received`, `account_id`, `device_id`, `device_tag`, `headers`, `body`)
                VALUES (%s, NOW(), %s, %s, %s, %s, %s)""".format(table), (
                    header['service'],
                    header.get('account_id', None),
                    header.get('device_id', None),
                    header.get('device_tag', None),
                    json.dumps(header),
                    stream.read(),
                )
            )
            self.conn.commit()

        elif header['type'] == 'download-file':
            try:
                file_id = header['file_id']
            except KeyError:
                raise StreamError("Invalid download file request, no file_id")
            if len(file_id) > 4096:
                raise StreamError("Invalid download file request, file_id is too long")

            file_path = self.save_stream_to_file(header, stream, self.file_location)

            cursor = self.conn.cursor()
            cursor.execute("""
                INSERT INTO file (`service`, `received`, `account_id`, `device_id`, `device_tag`, `headers`, `location`, `file_id`)
                VALUES (%s, NOW(), %s, %s, %s, %s, %s, %s)""", (
                    header['service'],
                    header.get('account_id', None),
                    header.get('device_id', None),
                    header.get('device_tag', None),
                    json.dumps(header),
                    file_path,
                    file_id,
                )
            )
            self.conn.commit()

        else:
            raise StreamError("Unrecognised header type {}".format(header['type']))

    @classmethod
    def save_stream_to_file(cls, header, stream, file_location):
        filename = cls.file_id_to_file_name(header['file_id'])
        path = os.path.join(
            file_location,
            header['service'],
            "{}".format(header.get('account_id', "None")),
            str(header.get('device_id', "None")),
            filename,
        )

        if len(path) > 250:
            raise StreamError("File path too long, unable to save stream.")

        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        if os.path.isfile(path):
            os.remove(path)

        with open(path, 'w') as f:
            shutil.copyfileobj(stream, f)

        return path

    @staticmethod
    def file_id_to_file_name(file_id):
        """
        Sometimes file ids are not the file names on the device, but are instead generated
        by the API. These are not guaranteed to be valid file names so need hashing.
        """
        if len(file_id) == 40 and re.match("^[a-f0-9]+$", file_id):
            return file_id
        # prefix with "re_" to avoid name collision with real fileids
        return "re_{}".format(hashlib.sha1(file_id).hexdigest())



from setuptools import setup
from pip.req import parse_requirements

from ricloud import __version__

PACKAGE_NAME = 'ricloud'


def markdown2rst(path):
    try:
        import pypandoc
        return pypandoc.convert(path, 'rst', 'md')
    except ImportError:
        with open(path, 'r') as f:
            return f.read()


def extract_requirements(path):
    requirements = parse_requirements(path, session=False)
    return [str(dep.req) for dep in requirements]


setup(
    name=PACKAGE_NAME,

    version=__version__,

    description="Python client for Reincubate's ricloud API.",
    long_description=markdown2rst('README.md'),

    url='https://github.com/reincubate/ricloud',

    author='Reincubate',
    author_email='enterprise@reincubate.com',

    license='AGPLv3',

    packages=[PACKAGE_NAME, ],
    package_data={PACKAGE_NAME: [PACKAGE_NAME + '/*']},
    include_package_data=True,

    classifiers=['Development Status :: 5 - Production/Stable',
                 'Environment :: Console',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: GNU General Public License (GPL)',
                 'Natural Language :: English',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python :: 2.6',
                 'Programming Language :: Python :: 2.7',
                 'Topic :: Utilities'],

    install_requires=extract_requirements('requirements.txt'),
)
